from collections.abc import Callable

from pydantic import ValidationInfo


def validate_identifier(value: str, types: tuple[str, ...]) -> str:
    t = value.split("--")[0]
    if t not in types:
        raise ValueError(f"{value} is not one of {types}")
    return value


def identifier_of_type(*args: str) -> Callable[..., str]:
    """
    Validates that an identifier is of a certain type or types.
    """

    def validator(value: str) -> str:
        return validate_identifier(value, args)

    return validator


def validate_bin_field(value: str, info: ValidationInfo) -> str:
    if info.field_name and not info.field_name.endswith("_bin"):
        raise ValueError("The property name MUST end with '_bin'.")
    return value


def validate_hex_field(value: str, info: ValidationInfo) -> str:
    if info.field_name and not info.field_name.endswith("_hex"):
        raise ValueError("The property name MUST end with '_hex'.")
    return value



def choose_one_hash(hashes: dict[str, str]) -> dict[str, str] | None:
    if "MD5" in hashes:
        return {"MD5": hashes["MD5"]}
    if "SHA-1" in hashes:
        return {"SHA-1": hashes["SHA-1"]}
    if "SHA-256" in hashes:
        return {"SHA-256": hashes["SHA-256"]}
    if "SHA-512" in hashes:
        return {"SHA-512": hashes["SHA-512"]}

    k = next(iter(hashes), None)
    if k is not None:
        return {k: hashes[k]}

    return None



from enum import Enum


class AccountType(Enum):
    facebook = "facebook"
    ldap = "ldap"
    nis = "nis"
    openid = "openid"
    radius = "radius"
    skype = "skype"
    tacacs = "tacacs"
    twitter = "twitter"
    unix = "unix"
    windows_local = "windows-local"
    windows_domain = "windows-domain"


class AttackMotivation(Enum):
    accidental = "accidental"
    coercion = "coercion"
    dominance = "dominance"
    ideology = "ideology"
    notoriety = "notoriety"
    organizational_gain = "organizational-gain"
    personal_gain = "personal-gain"
    personal_satisfaction = "personal-satisfaction"
    revenge = "revenge"
    unpredictable = "unpredictable"


class AttackResourceLevel(Enum):
    individual = "individual"
    club = "club"
    contest = "contest"
    team = "team"
    organization = "organization"
    government = "government"


class EncryptionAlgorithm(Enum):
    aes_256_gcm = "AES-256-GCM"
    chacha20_poly1305 = "ChaCha20-Poly1305"
    mime_type_indicated = "mime-type-indicated"


class ExtensionType(Enum):
    new_sdo = "new-sdo"
    new_sco = "new-sco"
    new_sro = "new-sro"
    property_extension = "property-extension"
    toplevel_property_extension = "toplevel-property-extension"


class GroupingContext(Enum):
    suspicious_activity = "suspicious-activity"
    malware_analysis = "malware-analysis"
    unspecified = "unspecified"


class HashingAlgorithm(Enum):
    md5 = "MD5"
    sha1 = "SHA-1"
    sha_256 = "SHA-256"
    sha_512 = "SHA-512"
    sha3_256 = "SHA3-256"
    sha3_512 = "SHA3-512"
    ssdeep = "SSDEEP"
    tlsh = "TLSH"


class IdentityClass(Enum):
    individual = "individual"
    group = "group"
    system = "system"
    organization = "organization"
    class_ = "class"
    unknown = "unknown"


class ImplementationLanguage(Enum):
    applescript = "applescript"
    bash = "bash"
    c = "c"
    cpp = "c++"
    csharp = "c#"
    go = "go"
    java = "java"
    javascript = "javascript"
    lua = "lua"
    objective_c = "objective-c"
    perl = "perl"
    php = "php"
    powershell = "powershell"
    python = "python"
    ruby = "ruby"
    scala = "scala"
    swift = "swift"
    typescript = "typescript"
    visual_basic = "visual-basic"
    x86_32 = "x86-32"
    x86_64 = "x86-64"


class IndicatorType(Enum):
    anomalous_activity = "anomalous-activity"
    anonymization = "anonymization"
    benign = "benign"
    compromised = "compromised"
    malicious_activity = "malicious-activity"
    attribution = "attribution"
    unknown = "unknown"


class IndustrySector(Enum):
    agriculture = "agriculture"
    aerospace = "aerospace"
    automotive = "automotive"
    chemical = "chemical"
    commercial = "commercial"
    communications = "communications"
    construction = "construction"
    defense = "defense"
    education = "education"
    energy = "energy"
    entertainment = "entertainment"
    financial_services = "financial-services"
    government = "government "
    emergency_services = "emergency-services"
    government_local = "government-local"
    government_national = "government-national"
    government_public_services = "government-public-services"
    government_regional = "government-regional"
    healthcare = "healthcare"
    hospitality_leisure = "hospitality-leisure"
    infrastructure = "infrastructure"
    dams = "dams"
    nuclear = "nuclear"
    water = "water"
    insurance = "insurance"
    manufacturing = "manufacturing"
    mining = "mining"
    non_profit = "non-profit"
    pharmaceuticals = "pharmaceuticals"
    retail = "retail"
    technology = "technology"
    telecommunications = "telecommunications"
    transportation = "transportation"
    utilities = "utilities"


class InfrastructureType(Enum):
    amplification = "amplification"
    anonymization = "anonymization"
    botnet = "botnet"
    command_and_control = "command-and-control"
    exfiltration = "exfiltration"
    hosting_malware = "hosting-malware"
    hosting_target_lists = "hosting-target-lists"
    phishing = "phishing"
    reconnaissance = "reconnaissance"
    staging = "staging"
    unknown = "unknown"


class MalwareResult(Enum):
    malicious = "malicious"
    suspicious = "suspicious"
    benign = "benign"
    unknown = "unknown"


class MalwareCapabilities(Enum):
    accesses_remote_machines = "accesses-remote-machines"
    anti_debugging = "anti-debugging"
    anti_disassembly = "anti-disassembly"
    anti_emulation = "anti-emulation"
    anti_memory_forensics = "anti-memory-forensics"
    anti_sandbox = "anti-sandbox"
    anti_vm = "anti-vm"
    captures_input_peripherals = "captures-input-peripherals"
    captures_output_peripherals = "captures-output-peripherals"
    captures_system_state_data = "captures-system-state-data"
    cleans_traces_of_infection = "cleans-traces-of-infection"
    commits_fraud = "commits-fraud"
    communicates_with_c2 = "communicates-with-c2"
    compromises_data_availability = "compromises-data-availability"
    compromises_data_integrity = "compromises-data-integrity"
    compromises_system_availability = "compromises-system-availability"
    controls_local_machine = "controls-local-machine"
    degrades_security_software = "degrades-security-software"
    degrades_system_updates = "degrades-system-updates"
    determines_c2_server = "determines-c2-server"
    emails_spam = "emails-spam"
    escalates_privileges = "escalates-privileges"
    evades_av = "evades-av"
    exfiltrates_data = "exfiltrates-data"
    fingerprints_host = "fingerprints-host"
    hides_artifacts = "hides-artifacts"
    hides_executing_code = "hides-executing-code"
    infects_files = "infects-files"
    infects_remote_machines = "infects-remote-machines"
    installs_other_components = "installs-other-components"
    persists_after_system_reboot = "persists-after-system-reboot"
    prevents_artifact_access = "prevents-artifact-access"
    prevents_artifact_deletion = "prevents-artifact-deletion"
    probes_network_environment = "probes-network-environment"
    self_modifies = "self-modifies"
    steals_authentication_credentials = "steals-authentication-credentials"
    violates_system_operational_integrity = "violates-system-operational-integrity"


class MalwareType(Enum):
    adware = "adware"
    backdoor = "backdoor"
    bot = "bot"
    bootkit = "bootkit"
    ddos = "ddos"
    downloader = "downloader"
    dropper = "dropper"
    exploit_kit = "exploit-kit"
    keylogger = "keylogger"
    ransomware = "ransomware"
    remote_access_trojan = "remote-access-trojan"
    resource_exploitation = "resource-exploitation"
    rogue_security_software = "rogue-security-software"
    rootkit = "rootkit"
    screen_capture = "screen-capture"
    spyware = "spyware"
    trojan = "trojan"
    unknown = "unknown"
    virus = "virus"
    webshell = "webshell"
    wiper = "wiper"
    worm = "worm"


class NetworkSocketAddressFamily(Enum):
    UNSPEC = "AF_UNSPEC"
    INET = "AF_INET"
    IPX = "AF_IPX"
    APPLETALK = "AF_APPLETALK"
    NETBIOS = "AF_NETBIOS"
    INET6 = "AF_INET6"
    IRDA = "AF_IRDA"
    BTH = "AF_BTH"


class NetworkSocketType(Enum):
    SOCK_STREAM = "SOCK_STREAM"
    AF_ISOCK_DGRAMNET = "AF_ISOCK_DGRAMNET"
    SOCK_RAW = "SOCK_RAW"
    SOCK_RDM = "SOCK_RDM"
    SOCK_SEQPACKET = "SOCK_SEQPACKET"


class OpinionEnum(Enum):
    strongly_disagree = "strongly-disagree"
    disagree = "disagree"
    neutral = "neutral"
    agree = "agree"
    strongly_agree = "strongly-agree"


class PatternType(Enum):
    stix = "stix"
    pcre = "pcre"
    sigma = "sigma"
    snort = "snort"
    suricata = "suricata"
    yara = "yara"


class ProcessorArchitecture(Enum):
    alpha = "alpha"
    arm = "arm"
    ia_64 = "ia-64"
    mips = "mips"
    powerpc = "powerpc"
    sparc = "sparc"
    x86 = "x86"
    x86_64 = "x86-64"


class Region(Enum):
    africa = "africa"
    eastern_africa = "eastern-africa"
    middle_africa = "middle-africa"
    northern_africa = "northern-africa"
    southern_africa = "southern-africa"
    western_africa = "western-africa"
    americas = "americas"
    caribbean = "caribbean"
    central_america = "central-america"
    latin_america_caribbean = "latin-america-caribbean"
    northern_america = "northern-america"
    south_america = "south-america"
    asia = "asia"
    central_asia = "central-asia"
    eastern_asia = "eastern-asia"
    southern_asia = "southern-asia"
    south_eastern_asia = "south-eastern-asia"
    western_asia = "western-asia"
    europe = "europe"
    eastern_europe = "eastern-europe"
    northern_europe = "northern-europe"
    southern_europe = "southern-europe"
    western_europe = "western-europe"
    oceania = "oceania"
    antarctica = "antarctica"
    australia_new_zealand = "australia-new-zealand"
    melanesia = "melanesia"
    micronesia = "micronesia"
    polynesia = "polynesia"


class ReportType(Enum):
    attack_pattern = "attack-pattern"
    campaign = "campaign"
    identity = "identity"
    indicator = "indicator"
    intrusion_set = "intrusion-set"
    malware = "malware"
    observed_data = "observed-data"
    threat_actor = "threat-actor"
    threat_report = "threat-report"
    tool = "tool"
    vulnerability = "vulnerability"


class ThreatActorType(Enum):
    activist = "activist"
    competitor = "competitor"
    crime_syndicate = "crime-syndicate"
    criminal = "criminal"
    hacker = "hacker"
    insider_accidental = "insider-accidental"
    insider_disgruntled = "insider-disgruntled"
    nation_state = "nation-state"
    sensationalist = "sensationalist"
    spy = "spy"
    terrorist = "terrorist"
    unknown = "unknown"


class ThreatActorRole(Enum):
    agent = "agent"
    director = "director"
    independent = "independent"
    infrastructure_architect = "infrastructure-architect"
    infrastructure_operator = "infrastructure-operator"
    malware_author = "malware-author"
    sponsor = "sponsor"


class ThreatActorSophistication(Enum):
    none = "none"
    minimal = "minimal"
    intermediate = "intermediate"
    advanced = "advanced"
    expert = "expert"
    innovator = "innovator"
    strategic = "strategic"


class ToolType(Enum):
    denial_of_service = "denial-of-service"
    exploitation = "exploitation"
    information_gathering = "information-gathering"
    network_capture = "network-capture"
    credential_exploitation = "credential-exploitation"
    remote_access = "remote-access"
    vulnerability_scanning = "vulnerability-scanning"
    unknown = "unknown"


class WindowsIntegrityLevel(Enum):
    low = "low"
    medium = "medium"
    high = "high"
    system = "system"


class WindowsPEBinary(Enum):
    dll = "dll"
    exe = "exe"
    sys = "sys"


class WindowsRegistryDatatype(Enum):
    NONE = "REG_NONE"
    SZ = "REG_SZ"
    EXPAND_SZ = "REG_EXPAND_SZ"
    BINARY = "REG_BINARY"
    DWORD = "REG_DWORD"
    DWORD_BIG_ENDIAN = "REG_DWORD_BIG_ENDIAN"
    DWORD_LITTLE_ENDIAN = "REG_DWORD_LITTLE_ENDIAN"
    LINK = "REG_LINK"
    MULTI_SZ = "REG_MULTI_SZ"
    RESOURCE_LIST = "REG_RESOURCE_LIST"
    FULL_RESOURCE_DESCRIPTION = "REG_FULL_RESOURCE_DESCRIPTION"
    RESOURCE_REQUIREMENTS_LIST = "REG_RESOURCE_REQUIREMENTS_LIST"
    QWORD = "REG_QWORD"
    INVALID_TYPE = "REG_INVALID_TYPE"


class WindowsServiceStartType(Enum):
    AUTO_START = "SERVICE_AUTO_START"
    BOOT_START = "SERVICE_BOOT_START"
    DEMAND_START = "SERVICE_DEMAND_START"
    DISABLED = "SERVICE_DISABLED"
    SYSTEM_ALERT = "SERVICE_SYSTEM_ALERT"


class WindowsServiceType(Enum):
    KERNEL_DRIVER = "SERVICE_KERNEL_DRIVER"
    FILE_SYSTEM_DRIVER = "SERVICE_FILE_SYSTEM_DRIVER"
    WIN32_OWN_PROCESS = "SERVICE_WIN32_OWN_PROCESS"
    WIN32_SHARE_PROCESS = "SERVICE_WIN32_SHARE_PROCESS"


class WindowsServiceStatus(Enum):
    CONTINUE_PENDING = "SERVICE_CONTINUE_PENDING"
    PAUSE_PENDING = "SERVICE_PAUSE_PENDING"
    PAUSED = "SERVICE_PAUSED"
    RUNNING = "SERVICE_RUNNING"
    START_PENDING = "SERVICE_START_PENDING"
    STOP_PENDING = "SERVICE_STOP_PENDING"
    STOPPED = "SERVICE_STOPPED"


class DNSRecord(Enum):
    A = "A"  # Address record : Returns a 32-bit IPv4 address, most commonly used to map hostnames to an IP address
    # of the host, but it is also used for DNSBLs, storing subnet masks in RFC 1101, etc.
    AAAA = "AAAA"  # IPv6 address record : Returns a 128-bit IPv6 address, most commonly used to map hostnames
    # to an IP address of the host.
    AFSDB = "AFSDB"  # AFS database record : Location of database servers of an AFS cell. This record is commonly
    # used by AFS clients to contact AFS cells outside their local domain.
    APL = "APL"  # Address Prefix List : Specify lists of address ranges, e.g. in CIDR format,
    # for various address families. Experimental.
    CAA = "CAA"  # Certification Authority Authorization : DNS Certification Authority Authorization,
    # constraining acceptable CAs for a host/domain
    CDNSKEY = "CDNSKEY"  # Child copy of DNSKEY record, for transfer to parent
    CDS = "CDS"  # Child DS : Child copy of DS record, for transfer to parent
    CERT = "CERT"  # Certificate record : Stores PKIX, SPKI, PGP, etc.
    CNAME = "CNAME"  # Canonical name record : Alias of one name to another: the DNS lookup will continue
    # by retrying the lookup with the new name.
    CSYNC = "CSYNC"  # Child-to-Parent Synchronization : Specify a synchronization mechanism between a child
    # and a parent DNS zone. Typical example is declaring the same NS records in the parent and the child zone
    DHCID = "DHCID"  # DHCP identifier : Used in conjunction with the FQDN option to DHCP
    DLV = "DLV"  # DNSSEC Lookaside Validation record : For publishing DNSSEC trust anchors outside of the
    # DNS delegation chain. Uses the same format as the DS record. RFC 5074 describes a way of using these records.
    DNAME = "DNAME"  # Delegation name record : Alias for a name and all its subnames, unlike CNAME, which is an alias
    # for only the exact name. Like a CNAME, the DNS lookup will continue by retrying the lookup with the new name.
    DNSKEY = "DNSKEY"  # DNS Key record : The key record used in DNSSEC. Uses the same format as the KEY record.
    DS = "DS"  # Delegation signer : The record used to identify the DNSSEC signing key of a delegated zone
    EUI48 = "EUI48"  # MAC address (EUI-48) : A 48-bit IEEE Extended Unique Identifier.
    EUI64 = "EUI64"  # MAC address (EUI-64) : A 64-bit IEEE Extended Unique Identifier.
    HINFO = "HINFO"  # Host Information : Providing Minimal-Sized Responses to DNS Queries That Have QTYPE=ANY
    HIP = "HIP"  # Host Identity Protocol : Method of separating the end-point identifier and locator roles of IPs.
    HTTPS = "HTTPS"  # HTTPS Binding : RR that improves performance for clients that need to resolve many resources
    # to access a domain. More info in this IETF Draft by DNSOP Working group and Akamai technologies.
    IPSECKEY = "IPSECKEY"  # IPsec Key : Key record that can be used with IPsec
    KEY = "KEY"  # Key record
    KX = "KX"  # Key Exchanger record : Used with some cryptographic systems (not including DNSSEC) to identify
    # a key management agent for the associated domain-name. Note that this has nothing to do with DNS Security.
    LOC = "LOC"  # Location record : Specifies a geographical location associated with a domain name
    MX = "MX"  # Mail exchange record : List of mail exchange servers that accept email for a domain
    NAPTR = "NAPTR"  # Naming Authority Pointer : Allows regular-expression-based rewriting of domain names
    # which can then be used as URIs, further domain names to lookups, etc.
    NS = "NS"  # Name server record : Delegates a DNS zone to use the given authoritative name servers
    NSEC = "NSEC"  # Next Secure record : Part of DNSSEC—used to prove a name does not exist.
    NSEC3 = "NSEC3"  # Next Secure record version 3 : An extension to DNSSEC that allows proof of nonexistence
    # for a name without permitting zonewalking
    NSEC3PARAM = "NSEC3PARAM"  # NSEC3 parameters : Parameter record for use with NSEC3
    OPENPGPKEY = "OPENPGPKEY"  # OpenPGP public key record : A DNS-based Authentication of Named Entities (DANE) method
    # for publishing and locating OpenPGP public keys in DNS for a specific email address.
    PTR = "PTR"  # PTR Resource Record [de] : Pointer to a canonical name. Unlike a CNAME, DNS processing stops and
    # just the name is returned. The most common use is for implementing reverse DNS lookups.
    RRSIG = "RRSIG"  # DNSSEC signature : Signature for a DNSSEC-secured record set. Same format as the SIG record.
    RP = "RP"  # Responsible Person : Information about the responsible person(s) for the domain.
    # Usually an email address with the @ replaced by a .
    SIG = "SIG"  # Signature
    SMIMEA = "SMIMEA"  # S/MIME cert association : Associates an S/MIME certificate with a domain name
    # for sender authentication.
    SOA = "SOA"  # Start of [a zone of] authority record : Specifies authoritative information about a DNS zone,
    # including the primary name server, the email of the domain administrator, the domain serial number,
    # and several timers relating to refreshing the zone.
    SRV = "SRV"  # Service locator : Generalized service location record, used for newer protocols instead of
    # creating protocol-specific records such as MX.
    SSHFP = "SSHFP"  # SSH Public Key Fingerprint : Resource record for publishing SSH public host key fingerprints
    # in the DNS, in order to aid in verifying the authenticity of the host.
    SVCB = "SVCB"  # Service Binding : RR that improves performance for clients that need to resolve many
    # resources to access a domain.
    TA = "TA"  # Trust Authorities : Part of a deployment proposal for DNSSEC without a signed DNS root.
    TKEY = "TKEY"  # Transaction Key record : A method of providing keying material to be used with TSIG
    # that is encrypted under the public key in an accompanying KEY RR.
    TLSA = "TLSA"  # TLSA certificate association : A record for DANE. The TLSA DNS resource record is used to
    # associate a TLS server certificate or public key with the domain name where the record is found.
    TSIG = "TSIG"  # Transaction Signature : Can be used to authenticate dynamic updates as coming from an approved
    # client, or to authenticate responses as coming from an approved recursive name server similar to DNSSEC.
    TXT = "TXT"  # Text record : Originally for arbitrary human-readable text in a DNS record. Often carries
    # machine-readable data, such as opportunistic encryption, Sender Policy Framework, DKIM, DMARC, DNS-SD, etc.
    URI = "URI"  # Uniform Resource Identifier : Can be used for publishing mappings from hostnames to URIs.
    ZONEMD = "ZONEMD"  # Message Digests for DNS Zones : Provides a cryptographic message digest over DNS zone data.


class IPUsage(Enum):
    parking = "parking"
    vpn = "vpn"
    tor = "tor"



from datetime import UTC, datetime
from enum import Enum
from typing import Literal, Self

from pydantic.functional_validators import model_validator

from stidantic.types import (
    Identifier,
    StixCore,
    StixMarking,
)


# 7.2.1.3 Statement Marking
class StatementMarking(StixCore):
    """
    The Statement marking type defines the representation of a textual marking statement (e.g., copyright, terms of use,
    etc.) in a definition. The value of the definition_type property MUST be statement when using this marking type.
    Statement markings are generally not machine-readable, and this specification does not define any behavior or
    actions based on their values.

    Content may be marked with multiple statements of use. In other words, the same content can be marked both with a
    statement saying "Copyright 2019" and a statement saying, "Terms of use are ..." and both statements apply.
    """

    # A Statement (e.g., copyright, terms of use) applied to the content marked by this marking definition.
    statement: str


# 7.2.1.4 TLP Marking
class TLPMarking(StixCore):
    """
    The TLP marking type defines how you would represent a Traffic Light Protocol (TLP) marking in a definition
    property. The value of the definition_type property MUST be tlp when using this marking type.
    """

    # The TLP level [TLP] of the content marked by this marking definition, as defined in this section.
    tlp: str


# 7.2.1 Marking Definition
class MarkingDefinition(StixMarking):
    """
    The marking-definition object represents a specific marking. Data markings typically represent handling or
    sharing requirements for data and are applied in the object_marking_refs and granular_markings properties on
    STIX Objects, which reference a list of IDs for marking-definition objects.

    Two marking definition types are defined in this specification: TLP, to capture TLP markings, and Statement,
    to capture text marking statements. In addition, it is expected that the FIRST Information Exchange Policy (IEP)
    will be included in a future version once a machine-usable specification for it has been defined.

    Unlike other STIX Objects, Marking Definition objects cannot be versioned because it would allow for indirect
    changes to the markings on a STIX Object. For example, if a Statement marking is changed from "Reuse Allowed" to
    "Reuse Prohibited", all STIX Objects marked with that Statement marking would effectively have an updated marking
    without being updated themselves. Instead, a new Statement marking with the new text should be created and the
    marked objects updated to point to the new marking.
    """

    type: Literal["marking-definition"] = "marking-definition"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Marking Definition.
    name: str | None = None
    # The definition_type property identifies the type of Marking Definition. The value of the definition_type property
    # SHOULD be one of the types defined in the subsections below: statement or tlp (see sections 7.2.1.3 and 7.2.1.4).
    # Any new marking definitions SHOULD be specified using the extension facility described in section 7.3.
    # If the extensions property is not present, this property MUST be present.
    definition_type: str | None = None
    # The definition property contains the marking object itself (e.g., the TLP marking as defined in section 7.2.1.4,
    # the Statement marking as defined in section 7.2.1.3).
    # Any new marking definitions SHOULD be specified using the extension facility described in section 7.3.
    # If the extensions property is not present, this property MUST be present.
    definition: TLPMarking | StatementMarking | None = None

    @model_validator(mode="after")
    def definition_if_no_extensions(self) -> Self:
        if not self.extensions and (not self.definition and not self.definition_type):
            raise ValueError(
                "If the extensions property is not present, definition_type and definition properties MUST be present."
            )
        return self


STIX_ZERO_DATE = datetime(2017, 1, 20, 00, 00, 00, 00, tzinfo=UTC)


class TLP(Enum):
    """
    The following standard marking definitions MUST be used to reference or represent TLP markings.
    Other instances of tlp-marking MUST NOT be used or created (the only instances of TLP marking definitions
    permitted are those defined here).
    """

    white = MarkingDefinition(
        id=Identifier("marking-definition--613f2e26-407d-48c7-9eca-b8e91df99dc9"),
        definition_type="tlp",
        definition=TLPMarking(tlp="white"),
        name="TLP:WHITE",
        created=STIX_ZERO_DATE,
    )
    green = MarkingDefinition(
        id=Identifier("marking-definition--34098fce-860f-48ae-8e50-ebd3cc5e41da"),
        definition_type="tlp",
        definition=TLPMarking(tlp="green"),
        name="TLP:GREEN",
        created=STIX_ZERO_DATE,
    )
    amber = MarkingDefinition(
        id=Identifier("marking-definition--f88d31f6-486f-44da-b317-01333bde0b82"),
        definition_type="tlp",
        definition=TLPMarking(tlp="amber"),
        name="TLP:AMBER",
        created=STIX_ZERO_DATE,
    )
    red = MarkingDefinition(
        id=Identifier("marking-definition--5e57c739-391a-4eb3-b6be-7d15ca92d5ed"),
        definition_type="tlp",
        definition=TLPMarking(tlp="red"),
        name="TLP:RED",
        created=STIX_ZERO_DATE,
    )



from typing import Annotated, Literal, Self

from pydantic import AfterValidator, Field
from pydantic.functional_validators import model_validator
from pydantic.types import PositiveInt

from stidantic.types import Identifier, StixRelationship, StixTimestamp, StixType
from stidantic.validators import identifier_of_type


# 5.1 Relationship
class Relationship(StixRelationship):
    """
    The Relationship object is used to link together two SDOs or SCOs in order to describe how they are related to
    each other. If SDOs and SCOs are considered "nodes" or "vertices" in the graph, the Relationship Objects (SROs)
    represent "edges".

    STIX defines many relationship types to link together SDOs and SCOs. These relationships are contained in the
    "Relationships" table under each SDO and SCO definition. Relationship types defined in the specification SHOULD
    be used to ensure consistency. An example of a specification-defined relationship is that an indicator indicates
    a campaign. That relationship type is listed in the Relationships section of the Indicator SDO definition.

    STIX also allows relationships from any SDO or SCO to any SDO or SCO that have not been defined in this
    specification. These relationships MAY use the related-to relationship type or MAY use a user-defined
    relationship type. As an example, a user might want to link malware directly to a tool.
    They can do so using related-to to say that the Malware is related to the Tool but not describe how,
    or they could use delivered-by (a user-defined name they determined) to indicate more detail.

    Note that some relationships in STIX may seem like "shortcuts". For example, an Indicator doesn't really detect
    a Campaign: it detects activity (Attack Patterns, Malware, Infrastructure, etc.) that are often used by that
    campaign. While some analysts might want all of the source data and think that shortcuts are misleading,
    in many cases it's helpful to provide just the key points (shortcuts) and leave out the low-level details.
    In other cases, the low-level analysis may not be known or sharable, while the high-level analysis is.
    For these reasons, relationships that might appear to be "shortcuts" are not excluded from STIX.
    """

    type: Literal["relationship"] = "relationship"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The name used to identify the type of Relationship. This value SHOULD be an exact value listed in the
    # relationships for the source and target SDO, but MAY be any string.
    relationship_type: StixType
    # A description that provides more details and context about the Relationship,
    # potentially including its purpose and its key characteristics.
    description: str | None = None
    # The id of the source (from) object. The value MUST be an ID reference to an SDO or SCO
    # (i.e., it cannot point to an SRO, Bundle, Language Content, or Marking Definition).
    source_ref: Identifier
    # The id of the target (to) object. The value MUST be an ID reference to an SDO or SCO
    # (i.e., it cannot point to an SRO, Bundle, Language Content, or Marking Definition).
    target_ref: Identifier
    # This optional timestamp represents the earliest time at which the Relationship between the objects exists.
    # If this property is a future timestamp, at the time the start_time property is defined,
    # then this represents an estimate by the producer of the intelligence of the earliest
    # time at which relationship will be asserted to be true.
    # If it is not specified, then the earliest time at which the relationship between the objects exists
    # is not defined.
    start_time: StixTimestamp | None = None
    # The latest time at which the Relationship between the objects exists. If this property is a future timestamp,
    # at the time the stop_time property is defined, then this represents an estimate by the producer of the
    # intelligence of the latest time at which relationship will be asserted to be true.
    # If stop_time is not specified, then the latest time at which the relationship between
    # the objects exists is either not known, not disclosed, or has no defined stop time.
    stop_time: StixTimestamp | None = None

    @model_validator(mode="after")
    def validate_start_stop_interval(self) -> Self:
        """
        If start_time and stop_time are both defined, then stop_time MUST be later than the start_time value.
        """
        if self.start_time and self.stop_time and self.start_time > self.stop_time:
            raise ValueError(
                "the stop_time property MUST be greater than or equal to the timestamp in the start_time property"
            )
        return self


# 5.2 Sighting
class Sighting(StixRelationship):
    """
    A Sighting denotes the belief that something in CTI (e.g., an indicator, malware, tool, threat actor, etc.)
    was seen. Sightings are used to track who and what are being targeted, how attacks are carried out, and to
    track trends in attack behavior.

    The Sighting relationship object is a special type of SRO; it is a relationship that contains extra properties
    not present on the Generic Relationship object. These extra properties are included to represent data specific
    to sighting relationships (e.g., count, representing how many times something was seen), but for other purposes
    a Sighting can be thought of as a Relationship with a name of "sighting-of". Sighting is captured as a relationship
    because you cannot have a sighting unless you have something that has been sighted.
    Sighting does not make sense without the relationship to what was sighted.

    Sighting relationships relate three aspects of the sighting:
    ●      What was sighted, such as the Indicator, Malware, Campaign, or other SDO (sighting_of_ref)
    ●      Who sighted it and/or where it was sighted, represented as an Identity (where_sighted_refs)
    ●      What was actually seen on systems and networks, represented as Observed Data (observed_data_refs)

    What was sighted is required; a sighting does not make sense unless you say what you saw. Who sighted it,
    where it was sighted, and what was actually seen are optional. In many cases it is not necessary to provide
    that level of detail in order to provide value.

    Sightings are used whenever any SDO has been "seen". In some cases, the object creator wishes to convey very little
    information about the sighting; the details might be sensitive, but the fact that they saw a malware instance or
    threat actor could still be very useful. In other cases, providing the details may be helpful or even necessary;
    saying exactly which of the 1000 IP addresses in an indicator were sighted is helpful when tracking which of those
    IPs is still malicious.

    Sighting is distinct from Observed Data in that Sighting is an intelligence assertion ("I saw this threat actor")
    while Observed Data is simply information ("I saw this file"). When you combine them by including the linked
    Observed Data (observed_data_refs) from a Sighting, you can say "I saw this file, and that makes me think I saw
    this threat actor".
    """

    type: Literal["sighting"] = "sighting"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A description that provides more details and context about the Sighting.
    description: str | None = None
    # The beginning of the time window during which the SDO referenced by the sighting_of_ref property was sighted.
    first_seen: StixTimestamp | None = None
    # The end of the time window during which the SDO referenced by the sighting_of_ref property was sighted.
    last_seen: StixTimestamp | None = None
    # If present, this MUST be an integer between 0 and 999,999,999 inclusive and represents the number of times the
    # SDO referenced by the sighting_of_ref property was sighted.
    # Observed Data has a similar property called number_observed, which refers to the number of times the data was
    # observed. These counts refer to different concepts and are distinct.
    # For example, a single sighting of a DDoS bot might have many millions of observations of the network traffic
    # that it generates. Thus, the Sighting count would be 1 (the bot was observed once) but the Observed Data
    # number_observed would be much higher.
    # As another example, a sighting with a count of 0 can be used to express that an indicator was not seen at all.
    count: PositiveInt | None = None
    # An ID reference to the SDO that was sighted (e.g., Indicator or Malware).
    # For example, if this is a Sighting of an Indicator, that Indicator’s ID would be the value of this property.
    # This property MUST reference only an SDO.
    sighting_of_ref: Identifier | None = None
    # A list of ID references to the Observed Data objects that contain the raw cyber data for this Sighting.
    # For example, a Sighting of an Indicator with an IP address could include the Observed Data for the
    # network connection that the Indicator was used to detect.
    observed_data_refs: list[Annotated[Identifier, AfterValidator(identifier_of_type("observed-data"))]] | None = None
    # A list of ID references to the Identity or Location objects describing the entities or types of entities
    # that saw the sighting.
    # Omitting the where_sighted_refs property does not imply that the sighting was seen by the object creator.
    # To indicate that the sighting was seen by the object creator, an Identity representing the object creator
    # should be listed in where_sighted_refs.
    where_sighted_refs: (
        list[Annotated[Identifier, AfterValidator(identifier_of_type("identity", "location"))]] | None
    ) = None
    # The summary property indicates whether the Sighting should be considered summary data.
    # Summary data is an aggregation of previous Sightings reports and should not be considered primary source data.
    # Default value is false.
    # WARN: Spec says it's a string, but description describes a boolean which defaults to false...
    summary: bool | str | None = None

    @model_validator(mode="after")
    def validate_first_last_interval(self) -> Self:
        """
        If this property and the first_seen property are both defined, then this property
        MUST be greater than or equal to the timestamp in the first_seen property.
        """
        if self.first_seen and self.last_seen and self.first_seen > self.last_seen:
            raise ValueError(
                "the last_seen property MUST be greater than or equal to the timestamp in the first_seen property"
            )
        return self


SROs = Annotated[(Relationship | Sighting), Field(discriminator="type")]



from typing import Annotated, Type, get_args  # pyright: ignore[reportDeprecated] # noqa: UP035

from pydantic import Field

from stidantic.extension import ExtensionDefinition
from stidantic.language import LanguageContent
from stidantic.marking import MarkingDefinition
from stidantic.sco import SCOs
from stidantic.sdo import SDOs
from stidantic.sro import SROs
from stidantic.types import ExtensionType, Identifier, StixCommon, StixCore, StixDomain, StixObservable


# 8. Stix Bundle
class StixBundle(StixCore):
    id: Identifier
    type: str = "bundle"
    objects: list[
        Annotated[
            (SROs | SDOs | SCOs | MarkingDefinition | LanguageContent | ExtensionDefinition),
            Field(discriminator="type"),
        ]
        | StixCommon
    ]

    @classmethod
    def register_new_object(
        cls,
        definition: ExtensionDefinition,  # ExtensionDefinition pyright: ignore[reportUnusedParameter]
        extension: Type[StixDomain | StixObservable],  # pyright: ignore[reportDeprecated]  # noqa: UP006
    ) -> None:
        if not (
            set(definition.extension_types)
            & {
                ExtensionType.new_sco.value,
                ExtensionType.new_sdo.value,
                ExtensionType.new_sro.value,
            }
        ):
            raise ValueError(
                "New object extension must contain new_sdo, new_sro or new_sco in the extension_types property"
            )
        annotation = cls.model_fields["objects"].annotation
        union, *_annotations = get_args(get_args(get_args(annotation)[0])[0])  # pyright: ignore[reportAny]
        cls.model_fields["objects"].annotation = list[
            Annotated[
                union | extension,
                Field(discriminator="type"),
            ]
            | StixCommon
        ]
        cls.__stix_extensions__[definition.id] = definition
        cls.model_rebuild(force=True)  # pyright: ignore[reportUnusedCallResult]
